use crate::{
    db::ItemSearchParams,
    models::{
        item::ItemSpecialData,
        task::{Task, TaskDetails, TaskMeta},
    },
    services::comms::broadcast_to_room,
    static_content::{
        dynzone::{dynzone_by_type, DynzoneType},
        possession_type::possession_data,
        room::room_map_by_code,
    },
    DResult,
};

use super::{
    get_player_item_or_fail, search_item_for_user, user_error, CommandHandlingError, UResult,
    UserError, UserVerb, UserVerbRef, VerbContext,
};
#[double]
use crate::db::DBTrans;
use ansi::ansi;
use async_trait::async_trait;
use chrono::{Duration, Utc};
use mockall_double::double;

async fn place_has_power(trans: &DBTrans, place_type: &str, place_code: &str) -> DResult<bool> {
    match place_type {
        "room" => room_map_by_code()
            .get(place_code)
            .map(|r| Ok(r.has_power))
            .unwrap_or(Ok(false)),
        "dynroom" => trans
            .find_item_by_type_code(place_type, place_code)
            .await?
            .and_then(|place| match place.special_data.as_ref() {
                Some(ItemSpecialData::DynroomData {
                    dynzone_code,
                    dynroom_code,
                }) => DynzoneType::from_str(&dynzone_code)
                    .and_then(|dzt| dynzone_by_type().get(&dzt))
                    .and_then(|dz| dz.dyn_rooms.get(dynroom_code.as_str()))
                    .map(|dr| Ok(dr.has_power)),
                _ => None,
            })
            .unwrap_or(Ok(false)),
        _ => Ok(false),
    }
}

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(
        self: &Self,
        ctx: &mut VerbContext,
        _verb: &str,
        remaining: &str,
    ) -> UResult<()> {
        let (word, remaining) = match remaining.split_once(" ") {
            None => user_error(ansi!("Try <bold>plug in<reset> something").to_owned())?,
            Some(v) => v,
        };
        if word.trim() != "in" {
            user_error(ansi!("Try <bold>plug in<reset> something").to_owned())?;
        }
        let player_item = get_player_item_or_fail(ctx).await?;
        if player_item.death_data.is_some() {
            user_error("Plugging metal things in can make you dead, but being dead doesn't let you plug things in.".to_owned())?;
        }
        let remaining = remaining.trim();
        let item = search_item_for_user(
            ctx,
            &ItemSearchParams {
                limit: 1,
                include_loc_contents: true,
                item_type_only: Some("possession"),
                ..ItemSearchParams::base(&player_item, &remaining)
            },
        )
        .await
        .map_err(|e| match e {
            CommandHandlingError::UserError(m) => CommandHandlingError::UserError(
                m + " I only looked in the room - try dropping an item first.",
            ),
            e => e,
        })?;
        let charge_data = match item
            .possession_type
            .as_ref()
            .and_then(|pt| possession_data().get(pt))
            .and_then(|pd| pd.charge_data.as_ref())
        {
            Some(v) if v.electric_recharge => v,
            _ => user_error("You can't recharge that!".to_owned())?,
        };

        let (place_type, place_code) = player_item
            .location
            .split_once("/")
            .ok_or_else(|| UserError("Bad location".to_owned()))?;
        if !place_has_power(&ctx.trans, place_type, place_code).await? {
            user_error("You can't find any power sockets to plug in to.".to_owned())?;
        }

        if item.charges >= charge_data.max_charges {
            user_error("You realise it's already fully charged.".to_owned())?;
        }

        let refstr = item.refstr();
        if ctx
            .trans
            .check_task_by_type_code("ChargeItem", &refstr)
            .await?
        {
            user_error("It's already charging!".to_owned())?;
        }
        ctx.trans
            .upsert_task(&Task {
                meta: TaskMeta {
                    task_code: refstr.clone(),
                    next_scheduled: Utc::now() + Duration::try_minutes(1).unwrap(),
                    ..Default::default()
                },
                details: TaskDetails::ChargeItem { item: refstr },
            })
            .await?;
        broadcast_to_room(
            &ctx.trans,
            &player_item.location,
            None,
            &format!(
                "{} plugs in {}.\n",
                &player_item.display_for_sentence(1, true),
                &item.display_for_sentence(1, false)
            ),
        )
        .await?;
        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
