#[double]
use crate::db::DBTrans;
use crate::{
    message_handler::user_commands::{user_error, UResult, VerbContext},
    models::{
        item::{Item, ItemFlag, ItemSpecialData},
        task::{Task, TaskDetails, TaskMeta},
    },
    regular_tasks::{TaskHandler, TaskRunContext},
    services::comms::broadcast_to_room,
    static_content::{possession_type::ChargeData, room::room_map_by_code},
    DResult,
};
use async_trait::async_trait;
use chrono::Utc;
use log::warn;
use mockall_double::double;
use once_cell::sync::OnceCell;
use std::time;

use super::{Describer, PossessionData, PossessionType, TurnToggleHandler};

pub struct LanternHandler {}

#[async_trait]
impl TurnToggleHandler for LanternHandler {
    async fn turn_cmd(
        &self,
        ctx: &mut VerbContext,
        player: &Item,
        what: &Item,
        turn_on: bool,
    ) -> UResult<()> {
        let mut item_mut = (*what).clone();
        let old_state = match item_mut.special_data.as_ref() {
            Some(ItemSpecialData::ToggleData { turned_on }) => *turned_on,
            _ => false,
        };
        if old_state == turn_on {
            user_error(format!(
                "It's already turned {}!",
                if turn_on { "on" } else { "off" }
            ))?;
        }
        if turn_on && item_mut.charges <= 0 {
            user_error("It won't turn on! It probably needs charging.".to_owned())?;
        }

        broadcast_to_room(
            &ctx.trans,
            &player.location,
            None,
            &format!(
                "{} flicks {} a lantern.\n",
                &player.display_for_sentence(1, true),
                if turn_on { "on" } else { "off" }
            ),
        )
        .await?;

        item_mut.special_data = Some(ItemSpecialData::ToggleData { turned_on: turn_on });
        if turn_on {
            ctx.trans
                .upsert_task(&Task {
                    meta: TaskMeta {
                        task_code: item_mut.refstr(),
                        next_scheduled: Utc::now() + chrono::TimeDelta::try_minutes(5).unwrap(),
                        ..Default::default()
                    },
                    details: TaskDetails::DischargeLight {
                        item: item_mut.refstr(),
                    },
                })
                .await?;
            item_mut.flags.push(ItemFlag::IlluminatesRoom);
            // To make it a bit harder to use well-timed periods of darkness to
            // avoid drawing any charge, use one charge to turn it on (but if they
            // had exactly one left, don't immediately turn it off).
            item_mut.charges -= 1;
        } else {
            ctx.trans
                .delete_task("DischargeLight", &item_mut.refstr())
                .await?;
            item_mut.flags.retain(|f| f != &ItemFlag::IlluminatesRoom);
        }
        ctx.trans.save_item_model(&item_mut).await?;

        Ok(())
    }
}

static LANTERN_HANDLER: LanternHandler = LanternHandler {};

#[derive(Clone)]
pub struct DischargeTaskHandler;
#[async_trait]
impl TaskHandler for DischargeTaskHandler {
    async fn do_task(&self, ctx: &mut TaskRunContext) -> DResult<Option<time::Duration>> {
        let item_ref = match &ctx.task.details {
            TaskDetails::DischargeLight { ref item } => item,
            t => {
                warn!("DischargeTaskHandler got unexpected task {:#?}", t);
                return Ok(None);
            }
        };
        let (item_type, item_code) = match item_ref.split_once("/") {
            Some(v) => v,
            None => {
                warn!("DischargeTaskHandler got unexpected item ref {}", item_ref);
                return Ok(None);
            }
        };
        let item = match ctx
            .trans
            .find_item_by_type_code(item_type, item_code)
            .await?
        {
            None => return Ok(None),
            Some(v) => v,
        };
        let mut item = (*item).clone();
        item.charges = item.charges.max(1) - 1;
        let lit_location = match item.location.split_once("/") {
            Some((loc_type, loc_code)) if loc_type == "player" || loc_type == "npc" => {
                match ctx.trans.find_item_by_type_code(loc_type, loc_code).await? {
                    None => item.location.clone(),
                    Some(pl) => pl.location.clone(),
                }
            }
            _ => item.location.clone(),
        };
        if item.charges == 1 {
            broadcast_to_room(
                &ctx.trans,
                &lit_location,
                None,
                &format!(
                    "{} dims noticeably as if it was running out of power.\n",
                    item.display_for_sentence(1, true)
                ),
            )
            .await?
        }
        if item.charges == 0 {
            broadcast_to_room(
                &ctx.trans,
                &lit_location,
                None,
                &format!(
                    "{} dims briefly before completely going out.\n",
                    item.display_for_sentence(1, true)
                ),
            )
            .await?;
            item.flags.retain(|fl| {
                fl != &ItemFlag::IlluminatesRoom && fl != &ItemFlag::IlluminatesAdjacentRoom
            });
        }
        ctx.trans.save_item_model(&item).await?;
        if item.charges == 0 {
            Ok(None)
        } else {
            Ok(Some(time::Duration::from_secs(300)))
        }
    }
}
pub static DISCHARGE_TASK: &(dyn TaskHandler + Sync + Send) = &DischargeTaskHandler;

pub struct GeigerCounterDescriber;

#[async_trait]
impl Describer for GeigerCounterDescriber {
    async fn describe_for(&self, _trans: &DBTrans, _item: &Item, player: &Item) -> UResult<String> {
        let (loc_type, loc_code) = match player.location.split_once("/") {
            None => return Ok("It doesn't currently show a reading.\n".to_owned()),
            Some(v) => v,
        };
        let level = if loc_type != "room" {
            0
        } else {
            room_map_by_code()
                .get(loc_code)
                .map(|r| r.environment.radiation)
                .unwrap_or(0)
        };
        Ok(format!(
            "It shows a reading of {} millisieverts/hour.\n",
            level * 360
        ))
    }
}

pub fn data() -> &'static Vec<(PossessionType, PossessionData)> {
    static D: OnceCell<Vec<(PossessionType, PossessionData)>> = OnceCell::new();
    &D.get_or_init(|| vec![
        (PossessionType::ElectricLantern, PossessionData {
            display: "electric lantern",
            aliases: vec!["lantern"],
            details: "A rechargeable electric lantern. It is made of yellow plastic, interspersed with banks of tiny flat white LEDs. It looks like it would illuminate a dark room well, although it probably isn't bright enough to let you see anything in the next room over.",
            weight: 300,
            turn_toggle_handler: Some(&LANTERN_HANDLER),
            charge_data: Some(ChargeData {
                max_charges: 20,
                charge_name_prefix: "bar",
                charge_name_suffix: "of power",
                electric_recharge: true,
                ..ChargeData::default()
            }),
            ..Default::default()
        }),
        (PossessionType::GeigerCounter, PossessionData {
            display: "geiger counter",
            aliases: vec!["counter"],
            details: "A sturdy black rectangular box with a yellow rubber protective case. On the front is a display showing numbers. Beneath the display is the text: 'OORANS approved Geiger Counter. Calibrated to predict effective dose of radiation from exposure to standard fallout. For use by trained personnel only'. ",
            weight: 300,
            computed_extra_details: Some(&GeigerCounterDescriber),
            ..Default::default()
        })
    ])
}
